#include "icp_dialogimpl.h"

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/DataStructures/src/NeighbourHood.h"

#include "../../../Utilities/UserMessages/src/StatusBar.h"

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>

#include <algorithm>
#include <functional>

#include <fstream>

// namespace std {
// 	template <typename T>
// 	T min(const T& a, const T& b) { return a < b ? a : b; }
// 	template <typename T>
// 	T max(const T& a, const T& b) { return a > b ? a : b; }
// }

/* 
 *  Constructs a ICP_Dialog which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 *
 *  The dialog will by default be modeless, unless you set 'modal' to
 *  TRUE to construct a modal dialog.
 */
ICP_Dialog::ICP_Dialog( QWidget* parent,  const char* name, bool modal, WFlags fl )
    : ICP_DialogBase( parent, name, modal, fl )
{
}

/*  
 *  Destroys the object and frees any allocated resources
 */
ICP_Dialog::~ICP_Dialog()
{
    // no need to delete child widgets, Qt does it all for us
}

ICP_Dialog *ICP_Dialog::getInstance() {
	if (dialog == 0) dialog = new ICP_Dialog();
	return dialog;
}

void ICP_Dialog::destroyInstance() {
	if (dialog != NULL) dialog->close();
	delete dialog;
	dialog = NULL;
}

void ICP_Dialog::handleRun()
{
	StatusBar *statusBar = StatusBar::getInstance();
	Scene *scene = Scene::getInstance();
	uint nofObj = scene->getNofObjects();

	SurfelMatrix sflMat(nofObj);
	NbHoodVector nbhVec(nofObj);

	assert(dialog);
	nofNeighbours = dialog->getNofNeighbours();
	nofIterations = dialog->getNofIterations();
	xi            = dialog->getXi();
	
	statusBar->resetProgress();
	statusBar->showMessage ("Building kD-Trees...", 0);

	// for each object, build object space kdtree
	this->buildKdTree(nbhVec, sflMat);

	objects.clear();
	for(Object *obj = scene->getFirstObject(); obj != NULL; obj = scene->getNextObject())
		objects.push_back(TrafoObject(obj));

	statusBar->showMessage ("Performing ICP...", 0);

	for(int i = 1; i <= nofIterations; ++i)
	{
		icpStep(nbhVec, sflMat);
		statusBar->setProgress (100.0 / nofIterations * i, true);
	}

	statusBar->showMessage ("", 1);
}

// build kdtree for all objects
void ICP_Dialog::buildKdTree(NbHoodVector& nbhVec, SurfelMatrix& sflMat)
{
	Scene *scene = Scene::getInstance();
	SurfelCollection *coll;
	uint i = 0;
	
	for(Object* obj = scene->getFirstObject();
		obj != NULL; obj = scene->getNextObject(), ++i)
	{
		coll = obj->getSurfelCollection();
		nbhVec[i] = coll->buildNeighbourHood();
		nbhVec[i]->setNofNeighbours(nofNeighbours);
		sflMat[i] = coll->getSurfelVector();
	}
}

// perform one icp step
void ICP_Dialog::icpStep(NbHoodVector& nbhVec, SurfelMatrix& sflMat)
{
	Scene* scene = Scene::getInstance();
	uint nofSfl, nofObj = scene->getNofObjects();

	// query surfel
	SurfelInterface*  qSfl;
	uint              qSflIdx;

	// query collection and object
	SurfelCollection* qColl;
	Objects::iterator qObj;
	uint              qObjIdx;

	// other object
	Objects::iterator oObj;
	uint              oObjIdx;

	// query point in other local coordinates
	Vector3D qLocalO, nbWorld;

	// neighbour indices and distances
	uint  nbIdx, nbPosIdx, minNbIdx;
	float nbDistSq, dMax = 0, sigmaSqInv;
	float weight, weightSum, distance;

	gsl_matrix* uMat    = gsl_matrix_alloc(nofNeighbours, 3);
	gsl_matrix* vMat    = gsl_matrix_alloc(3, 3);

	gsl_vector* weightVec = gsl_vector_alloc(nofNeighbours);
	gsl_vector* cogVec    = gsl_vector_alloc(3);
	gsl_vector* sVec      = gsl_vector_alloc(3);
	gsl_vector* workVec   = gsl_vector_alloc(3);

	// index queue for closest point
	MaxPriorityQueue<Correspondence, float> cpQueue; cpQueue.setSize(nofNeighbours);
	MaxPriorityQueue<Correspondence, float>::Element* elPtr;
	
	CorrespondenceVector corrVec;
	PointVector          qWorldVec;
	Vector3D             qWorld;

	int i;

	// iterate objects
	for(qObj = objects.begin(), qObjIdx = 0; qObjIdx < nofObj; ++qObj, ++qObjIdx)
	{
		qColl = qObj->obj->getSurfelCollection();
		nofSfl    = qColl->getNofSurfels();

		corrVec.clear();
		qWorldVec.resize(nofSfl);

		// iterate surfels in object and
		// find closest neighbours ...
		for(qSfl = qColl->getFirstSurfel(), qSflIdx = 0;
			qSfl != NULL; qSfl = qColl->getNextSurfel(), ++qSflIdx)
		{

			// transform query point to world space
			MtrMultVector3DHom(qObj->trf, qSfl->getPosition(), &qWorld);
			qWorldVec[qSflIdx] = qWorld;

			cpQueue.init();

			// ... in all objects except the current one
			for(oObj = objects.begin(), oObjIdx = 0; oObjIdx < nofObj; ++oObj, ++oObjIdx)
			{
				if(qObj == oObj) continue;

				// transform query point to otherObj space
				MtrMultVector3DHom(oObj->inv, qWorld, &qLocalO);

				// perform kd-tree search
				if(cpQueue.isEmpty()) nbhVec[oObjIdx]->clearMaxQueryDistance();
				else nbhVec[oObjIdx]->setMaxQuerySqrDistance( dMax );
				nbhVec[oObjIdx]->setSourcePoint(qLocalO);	

				// add points to closest point queue if they are closer
				for(uint nbIdx = 0; nbIdx < nbhVec[oObjIdx]->getNofFoundNeighbours(); ++nbIdx)
				{
					nbPosIdx = nbhVec[oObjIdx]->getNeighbourPositionIndex(nbIdx);
					nbDistSq = nbhVec[oObjIdx]->getSquaredDistance(nbIdx);
					if(nbDistSq < dMax || !cpQueue.isFull()) 
					{
						cpQueue.insert(Correspondence(oObjIdx, sflMat.at(oObjIdx)->at(nbPosIdx)->getPosition(), 0), nbDistSq);
						dMax = cpQueue.getMaxWeight();
					}
				}
			}

			assert(cpQueue.isFull());
	
			// find the minimum squered distance
			nbDistSq = 100000;
			for(elPtr = cpQueue.getQueue(), nbIdx = 0; nbIdx < unsigned(nofNeighbours); ++elPtr, ++nbIdx)
			{
				if(elPtr->weight < nbDistSq) 
				{
					minNbIdx = nbIdx;
					nbDistSq = elPtr->weight;
				}
			}

			// get the distance between furthest and closest point
			distance = (cpQueue.getQueue()->index.pnt - (cpQueue.getQueue() + minNbIdx)->index.pnt).getSquaredLength();

			// check if furthest and closest point are significantly further away than the 
			// closest point and the query point.
			if(nbDistSq * nofNeighbours > distance)
			{ // just take the closest point

				// just take the closest point
				elPtr = cpQueue.getQueue() + minNbIdx;

				// transform nearest point to world space
				MtrMultVector3DHom(objects[elPtr->index.idx].trf, elPtr->index.pnt, &nbWorld);

				// add corresponding point to array (in world space)
				corrVec.push_back(Correspondence(qSflIdx, nbWorld, (qWorld - nbWorld).getLength()));

			} 
			else
			{ // do weighted least squares plane fit

				// set sigma so that the minimum weight is ~0.01
				// the furtherst point is in front of the queue
				sigmaSqInv = 4.5f / cpQueue.getQueue()->weight;

				// transform nearest neighbour positions to world space 
				// and compute weighted center of gravity
				gsl_vector_set_zero(cogVec);
				weightSum = 0;
				for(elPtr = cpQueue.getQueue(), nbIdx = 0; nbIdx < unsigned(nofNeighbours); ++elPtr, ++nbIdx)
				{
					// transform nearest neighbour to world space
					MtrMultVector3DHom(objects[elPtr->index.idx].trf, elPtr->index.pnt, &nbWorld);
					// compute weight
					weight = exp(-elPtr->weight * sigmaSqInv);
	
					gsl_vector_set(weightVec, nbIdx, weight);
					for(i = 0; i < 3; ++i)
					{
						nbWorld[i] *= weight;
						gsl_matrix_set(uMat, nbIdx, i, nbWorld[i]);
						*gsl_vector_ptr(cogVec, i)  += nbWorld[i];
					}
					weightSum += weight;
				}


				// scale cog by 1 / weightSum
				gsl_blas_dscal(1.0f / weightSum, cogVec);
			
				// compute the rank-1 update U = w cog^T + U.
				gsl_blas_dger (-1.0, weightVec, cogVec, uMat);
	
				// compute singular value decomposition (weightVec is undefined afterwards)
				gsl_linalg_SV_decomp(uMat, vMat, sVec, workVec);
	
				// compute the closest point and the distance
				distance = 0;
				for(i = 0; i < 3; ++i)
				{
					distance += gsl_matrix_get(vMat, i, 2) * (gsl_vector_get(cogVec, i) - qWorld[i]);
				}
				for(i = 0; i < 3; ++i)
					qWorld[i] += distance * gsl_matrix_get(vMat, i, 2);

				// add corresponding point to array (in world space)
				corrVec.push_back(Correspondence(qSflIdx, qWorld, fabsf(distance)));
			}

		}

		// find the xi closest points (-> trimmed icp)
		nofSfl *= xi;
		std::nth_element(corrVec.begin(), corrVec.begin() + nofSfl, corrVec.end());
		corrVec.resize(nofSfl);
/*
		// begin debug:
		for(int i = 0; i < nofSfl; ++i)
		{
			SurfelCollection* sflColl = qObj->obj->getSurfelCollection();

			// color the points that have a close neighbour
			SurfelInterface *corrSfl = sflColl->getFirstSurfel();
			for(int count = 0; count < corrVec[i].idx; ++count) corrSfl = sflColl->getNextSurfel();
			scene->setSurfelDiffuseColor(corrSfl, 0xFF0000FF, false);

			// add corresponding point to object
			SurfelInterface *surfel = sflColl->addSurfel(corrSfl, false);
			// move point from world to local coordinates
			MtrMultVector3DHom(qObj->inv, corrVec[i].pnt, &nbWorld);
			scene->setSurfelPosition(surfel, nbWorld, false);
			scene->setSurfelDiffuseColor(surfel, 0xFFFF0000, false);
		}

		Scene::getInstance()->emitRenderFrameSignal();
		break;
		// end debug
*/		
		// perform transformation step
		transform(qWorldVec, corrVec, &(*qObj));

		// redraw
		Scene::getInstance()->emitRenderFrameSignal();
	}

	gsl_matrix_free(uMat);
	gsl_matrix_free(vMat);

	gsl_vector_free(weightVec);
	gsl_vector_free(cogVec);
	gsl_vector_free(sVec);
	gsl_vector_free(workVec);

}

// implements svd-algo from 'estimating 3-d rigid body transofrmations' by D.W. Eggert
void ICP_Dialog::transform(PointVector& qWorldVec, CorrespondenceVector& corrVec, TrafoObject* qObj)
{

#if 0 // global registration (in contrast to stepwise)
	std::swap(qWorldVec, PointVector());
	SurfelCollection* coll = qObj->obj->getSurfelCollection();
	SurfelInterface* sfl;
	for(sfl = coll->getFirstSurfel(); sfl != NULL; sfl = coll->getNextSurfel())
		qWorldVec.push_back(sfl->getPosition());
#endif

	Correspondence *corr;

	uint corrIdx;
	uint nofCorr = corrVec.size();
	float scale = 1.0 / nofCorr;

	int i, j, k;

	Vector3D qPnt, cPnt, qAvg(0.0, 0.0, 0.0), cAvg(0.0, 0.0, 0.0);

	gsl_matrix* hMat    = gsl_matrix_alloc(3, 3);
	gsl_matrix* vMat    = gsl_matrix_alloc(3, 3);

	gsl_vector* sVec    = gsl_vector_alloc(3);
	gsl_vector* workVec = gsl_vector_alloc(3);

	Trafo rotMat = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1};
	Trafo rotOldMat, tmpMat;

	// find median of obj- and corresponding-points
	for(corrIdx = 0; corrIdx < nofCorr; ++corrIdx)
	{
		corr = &corrVec[corrIdx];
		qAvg += qWorldVec[corr->idx];
		cAvg += corr->pnt;
	}
	for(i = 0; i < 3; ++i) 
	{
		qAvg[i] *= scale; 
		cAvg[i] *= scale; 
	}

	// build H-matrix
	gsl_matrix_set_zero(hMat);
	for(corrIdx = 0; corrIdx < nofCorr; ++corrIdx)
	{
		corr = &corrVec[corrIdx];
		qPnt = qWorldVec[corr->idx] - qAvg;
		cPnt = corr->pnt - cAvg;

		for(i = 0; i < 3; ++i)
			for(j = 0; j < 3; ++j)
				*gsl_matrix_ptr(hMat, i, j) += qPnt[i] * cPnt[j];
	}

	gsl_linalg_SV_decomp(hMat, vMat, sVec, workVec);

	// compute rotation R = V*U'
	for(i = 0; i < 3; ++i) // column
	{
		for(j = 0; j < 3; ++j) // row
		{
			float t = 0;
			for(k = 0; k < 3; ++k)
				t += gsl_matrix_get(vMat, i, k) * gsl_matrix_get(hMat, j, k);
			rotMat[j*4 + i] = t; // column major
		}

	}

/*	std::ofstream file("result.txt", std::ios_base::out | std::ios_base::app);
	file << "R =\n";
	for(i = 0; i < 4; ++i)
	{
		for(j = 0; j < 4; ++j) file << rotMat[j*4 + i] << " ";
		file << std::endl;
	}
	file << std::endl;
*/
	// apply rotation, 
	qObj->obj->getRotationMatrix(rotOldMat);
	MtrMult4x4f(rotMat, rotOldMat, tmpMat);
	qObj->obj->setRotationMatrix(tmpMat, false);

	// get old translation vector - qAvg
	qObj->obj->getTranslationMatrix(tmpMat);
	for(i = 0; i < 3; ++i) qPnt[i] = tmpMat[12 + i] - qAvg[i];

	// multiply with new rotation and add cAvg
	MtrMultVector3DHom(rotMat, qPnt, &cPnt);
	for(i = 0; i < 3; ++i) cPnt[i] += cAvg[i];

	// apply translation
	qObj->obj->setPosition(cPnt, false);

	// update transformation.
	qObj->updateTrafo();

	gsl_matrix_free(hMat);
	gsl_matrix_free(vMat);

	gsl_vector_free(sVec);
	gsl_vector_free(workVec);

}


ICP_Dialog* ICP_Dialog::dialog  = NULL;
