#ifndef ICP_DIALOG_H
#define ICP_DIALOG_H

#include "icp_dialog.h"
#include <qspinbox.h> 

#include <vector>

#include "../../../Core/Scene/src/Object.h"
#include "../../../Utilities/src/Matrix.h"
class Neighbourhood;

class ICP_Dialog : public ICP_DialogBase
{ 

    Q_OBJECT
	static ICP_Dialog* dialog;
	
	typedef MyDataTypes::TransformationMatrix16f Trafo;

	// surfels pointer vector
	typedef std::vector<SurfelInterface*>    SurfelVector;
	// surfels pointer matrix (dim2 = object, dim1 = surfel)	
	typedef std::vector<SurfelVector*>  SurfelMatrix;
	// neighbourhood pointer vector
	typedef std::vector<NeighbourHood*> NbHoodVector;
	// vector of points
	typedef std::vector<Vector3D>       PointVector;

	// a surfel idx and its closest point in space
	struct Correspondence {
		Correspondence() {}
		Correspondence(uint i, const Vector3D& p, float d) : idx(i), pnt(p), dist(d) {}
		bool operator<(const Correspondence& c) const { return dist < c.dist; }
		// id of the surfel collection
		uint     idx;
		// local position of the point
		Vector3D pnt;
		// local distance to the query point
		float    dist;
	};

	// vector of correspondences
	typedef std::vector<Correspondence> CorrespondenceVector;

	// an object and its transformation matrix plus inverse
	struct TrafoObject {
		TrafoObject(Object* o) : obj(o)
		{
			updateTrafo();
		}
		void updateTrafo()
		{
			obj->getTransformationMatrix(trf);
			MtrInverse4x4f(trf, inv);
		}

		Object* obj;
		Trafo   trf, inv;
	};

	typedef std::vector<TrafoObject> Objects;
	Objects objects;

	int   nofNeighbours;
	int   nofIterations;
	float xi;

public:

	/**
	 * This method returns the number of points a plane should be fitted through
	 * during corresponding point search.
	 *
	 * @return the number of points a plane should be fitted through.
	 */
	int   getNofNeighbours() { return dynamic_cast<ICP_DialogBase*>(dialog)->PointsSpin->value(); };
	int   getNofIterations() { return dynamic_cast<ICP_DialogBase*>(dialog)->IterSpin->value(); }
	float getXi() { return 0.01f * dynamic_cast<ICP_DialogBase*>(dialog)->TrimSpin->value(); }

	/**
	 * This method creates an instance of this <code>ICP_Dialog</code>, if necessary,
	 * and returns it. Only one instance at a time exists (singleton pattern).
	 *
	 * @return an instance of this <code>Sampling_DialogImpl</code>
	 */
	static ICP_Dialog *getInstance();

	/**
	 * Destroys the instance of this <code>Sampling_DialogImpl</code>, if there
	 * is one.
	 */
	static void destroyInstance();
	
protected:  // To suppress warnings...

	virtual ~ICP_Dialog();

private:

    ICP_Dialog( QWidget* parent = 0, const char* name = 0, bool modal = FALSE, WFlags fl = 0 );
    
	// Computes sqrt(a^2 + b^2) without destructive underflow or overflow.
	float pythag(float a, float b)
	{
		float aa = fabs(a);
		float ab = fabs(b);
		if (aa > ab) return aa * sqrt(1.0 + SQR(ab / aa));
		else return (ab == 0.0 ? 0.0 : ab * sqrt(1.0+SQR(aa/ab)));
	}

	void buildKdTree(NbHoodVector&, SurfelMatrix&);
	void icpStep(NbHoodVector&, SurfelMatrix&);
	void transform(PointVector&, CorrespondenceVector&, TrafoObject*);

private slots:
	
	virtual void handleRun();

};

#endif // ICP_DIALOG_H
